from .base import (
    ObservationData as ObservationData,
    ResponseData as ResponseData,
    SpectrumData as SpectrumData,
)
from .ogip import (
    Data as Data,
    Response as Response,
    Spectrum as Spectrum,
)
